#include "apr_tables.h"
#include "apr_strings.h"
#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_request.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_main.h"
#include "mod_cspace.h"
#include "cspace_utils.h"

module AP_MODULE_DECLARE_DATA cspace_module;

typedef struct cspace_cfg {
    apr_pool_t *pool;
    apr_array_header_t *use_cspace; /*TODO: use an array to support many pages*/ 
                                    /*how about using a hash table??*/
    char *key_file;
    char dirspec[256]; /*TODO: use malloc?*/
    char *epr;
} cspace_cfg;

#define dbg_printf printf

#ifdef CSPACE_DEBUG
#define cfg_printf(st, cfg, nl) cfg_printf_fn((st), (cfg), (nl))
static void cfg_printf_fn(const char* st, cspace_cfg *cfg, const char* nl) {
    printf("%s: %p, use_cspace: %s size:%d, dir: %s key:%s epr:%s\n%s", st, (void*)cfg, 
           cfg->use_cspace?"true":"false", cfg->use_cspace->nelts, 
           cfg->dirspec, cfg->key_file, cfg->epr, nl);
}
#else
#define cfg_printf(st, cfg, nl) 
#endif

static const char *cmd_use_cspace(cmd_parms *cmd, void *mconfig, 
                        const char *arg) {
    const char *path;
#ifdef CSPACE_DEBUG
    dbg_printf("cmd_use_cspace: arg:%s\n", arg);
#endif    
    /*mconfig is null for per server config*/
    /*
     * in case of a per dir config, we would have done something like:
     *      cspace_cfg *cfg = (cspace_cfg *) mconfig;
     *      if (mconfig) {
     *          cfg->use_cspace = (char *) arg;
     *          cfg_printf("\tparsed cfg", cfg, "\n");
     *      }
     */
    
    cspace_cfg *cfg = ap_get_module_config(cmd->server->module_config, 
            &cspace_module);
    
    /*cfg->use_cspace = (char *)arg;*/
    path = (const char *)arg;
    
    *(const char **)apr_array_push((cfg->use_cspace)) = 
        apr_pstrdup(cfg->pool, path);

    /* *(const char **)apr_array_push(cfg->use_cspace) = 
        apr_pstrdup(cmd->server->process->pool, path);*/

#ifdef CSPACE_DEBUG
    cfg_printf("\tuse_cspace cfg", cfg, "\n");
#endif
    return NULL;
}

static const char *cmd_key_file(cmd_parms *cmd, void *mconfig, 
                        const char *arg) {
#ifdef CSPACE_DEBUG
    dbg_printf("cmd_key_file: arg:%s\n", arg);
#endif

    cspace_cfg *cfg = ap_get_module_config(cmd->server->module_config,
            &cspace_module);
    cfg->key_file = (char *)arg;

#ifdef CSPACE_DEBUG
    cfg_printf("\tkey_file cfg", cfg, "\n");
#endif
   
    return NULL;
}

static const char *cmd_cspace_epr(cmd_parms *cmd, void *mconfig, 
                        const char *arg) {
#ifdef CSPACE_DEBUG
    dbg_printf("cmd_cspace_epr: arg:%s\n", arg);
#endif

    cspace_cfg *cfg = ap_get_module_config(cmd->server->module_config,
            &cspace_module);
    cfg->epr = (char *)arg;

#ifdef CSPACE_DEBUG
    cfg_printf("\tkey_file epr", cfg, "\n");
#endif
   
    return NULL;
}

static const command_rec cspace_cmds[] = {
    AP_INIT_ITERATE("UseCSpaceFor", cmd_use_cspace, NULL, RSRC_CONF, 
            "Path to the file Cardspace should be used"),
    AP_INIT_TAKE1("CSpaceTokenVerifierService", cmd_cspace_epr, NULL, RSRC_CONF, 
            "EPR of relying party service"),
    AP_INIT_TAKE1("SSLCertificateKeyFile", cmd_key_file, NULL, RSRC_CONF,
            "SSL Server Private Key file"),
    {NULL}
};

static void *cspace_cfg_persvr_merge(apr_pool_t *p, void *parent_cfg, 
                    void *own_cfg) {
#ifdef CSPACE_DEBUG
    dbg_printf("cspace_cfg_persvr_merge: \n");
#endif
   
    cspace_cfg *owncfg = (cspace_cfg *)own_cfg; 
    cspace_cfg *pcfg = (cspace_cfg *)parent_cfg;
    cspace_cfg *merged_cfg = (cspace_cfg *)apr_palloc(p, sizeof(cspace_cfg));
    
    if (pcfg) {
        if (!apr_is_empty_array(pcfg->use_cspace)) 
            merged_cfg->use_cspace = pcfg->use_cspace;
        if (pcfg->key_file) merged_cfg->key_file = pcfg->key_file;
        if (pcfg->epr) merged_cfg->epr = pcfg->epr;
    }
    if (owncfg) {
        if (!apr_is_empty_array(owncfg->use_cspace))
            merged_cfg->use_cspace = owncfg->use_cspace;
        if (owncfg->key_file) merged_cfg->key_file = owncfg->key_file;
        if (owncfg->epr) merged_cfg->epr = owncfg->epr;
    }

#ifdef CSPACE_DEBUG
    cfg_printf("\tparent", pcfg, "\n");
    cfg_printf("\town", owncfg, "\n");
    cfg_printf("\tmeerged", merged_cfg, "\n");
#endif

    
    return merged_cfg;
}


static void *cspace_cfg_persvr_create(apr_pool_t *p, server_rec *s) {
#ifdef CSPACE_DEBUG
    dbg_printf("cspace_cfg_persvr_create: \n");
#endif
   
    cspace_cfg *cfg = (cspace_cfg *)apr_palloc(p, sizeof(cspace_cfg));
    
    cfg->pool = p; /*TODO: create a sub pool*/

    cfg->dirspec[0] = '\0';
    cfg->key_file = NULL;

    /*cfg->use_cspace = NULL;*/
    cfg->use_cspace = apr_array_make(p, 1, sizeof(char *));
    /*what is "1" here?*/
            
    cfg->epr = NULL;

#ifdef CSPACE_DEBUG
    cfg_printf("\tserver cfg:", cfg, "\n");
#endif

    return (void *)cfg;
}

/*static void *cspace_cfg_perdir_create(apr_pool_t *p, char *dirspec) {
    //if (!dirspec) return NULL;
#ifdef CSPACE_DEBUG
    dbg_printf("cspace_cfg_perdir_create: \n");
#endif

    cspace_cfg *cfg = (cspace_cfg *)apr_palloc(p, sizeof(cspace_cfg));
    cfg->dirspec[0]='\0';
    cfg->key_file = NULL;
    cfg->use_cspace = NULL; 
    cfg->epr = NULL;
    if (dirspec) {
        strcpy(cfg->dirspec, dirspec);
    }

#ifdef CSPACE_DEBUG
    cfg_printf("\tcreated cfg", cfg, "\n");
#endif
   
    return (void *)cfg;
}*/

/*static void *cspace_cfg_perdir_merge(apr_pool_t *p, void *parent_cfg, 
        void *own_cfg) {
#ifdef CSPACE_DEBUG
    dbg_printf("cspace_cfg_perdir_merge:\n");
#endif

    cspace_cfg *merged_cfg = (cspace_cfg *) apr_palloc(p, sizeof(cspace_cfg));
    cspace_cfg *pcfg = (cspace_cfg *) parent_cfg;
    cspace_cfg *owncfg = (cspace_cfg *) own_cfg;

#ifdef CSPACE_DEBUG
    cfg_printf("\tparent cfg", pcfg, "");
    cfg_printf("\town    cfg", owncfg, "");
#endif

    strcpy(merged_cfg->use_cspace, pcfg->use_cspace); 
    strcpy(merged_cfg->dirspec, owncfg->dirspec);
    strcpy(merged_cfg->epr, owncfg->epr);

#ifdef CSPACE_DEBUG
    cfg_printf("\tmerged cfg", merged_cfg, "\n");
#endif
   
    return (void *)merged_cfg;
}*/

/* FROM http_filter.c:
 *
 * Here we deal with getting the request message body from the client.
 * Whether or not the request contains a body is signaled by the presence
 * of a non-zero Content-Length or by a Transfer-Encoding: chunked.
 *
 * Note that this is more complicated than it was in Apache 1.1 and prior
 * versions, because chunked support means that the module does less.
 *
 * The proper procedure is this:
 *
 * 1. Call ap_setup_client_block() near the beginning of the request
 *    handler. This will set up all the necessary properties, and will
 *    return either OK, or an error code. If the latter, the module should
 *    return that error code. The second parameter selects the policy to
 *    apply if the request message indicates a body, and how a chunked
 *    transfer-coding should be interpreted. Choose one of
 *
 *    REQUEST_NO_BODY          Send 413 error if message has any body
 *    REQUEST_CHUNKED_ERROR    Send 411 error if body without Content-Length
 *    REQUEST_CHUNKED_DECHUNK  If chunked, remove the chunks for me.
 *    REQUEST_CHUNKED_PASS     If chunked, pass the chunk headers with body.
 *
 *    In order to use the last two options, the caller MUST provide a buffer
 *    large enough to hold a chunk-size line, including any extensions.
 *
 * 2. When you are ready to read a body (if any), call ap_should_client_block().
 *    This will tell the module whether or not to read input. If it is 0,
 *    the module should assume that there is no message body to read.
 *
 * 3. Finally, call ap_get_client_block in a loop. Pass it a buffer and its size.
 *    It will put data into the buffer (not necessarily a full buffer), and
 *    return the length of the input block. When it is done reading, it will
 *    return 0 if EOF, or -1 if there was an error.
 *    If an error occurs on input, we force an end to keepalive.
 *
 *    This step also sends a 100 Continue response to HTTP/1.1 clients if appropriate.
 */

static int cspace_read_request(request_rec *r, char **buf, int *retval) {
    apr_array_header_t *a;
    apr_pool_t *p = r->pool;
    apr_pool_t *temp_pool;
    int res, ret;
    
#ifdef CSPACE_DEBUG
    dbg_printf("called: cspace_read_request ******************\n");
#endif

    if (apr_pool_create(&temp_pool, p) != APR_SUCCESS)
        return FAIL; /*TODO: should it be UNAUTHORIZED?*/
    
    a = apr_array_make(temp_pool, 1, sizeof(char *));
    
    ret = ap_setup_client_block(r, REQUEST_CHUNKED_DECHUNK);

    if (ret != OK) {
        *retval = ret;
        return FAIL; /*above: module should return that error code*/
    }
    
    if (ap_should_client_block(r)) {
        char temp_buf[HUGE_STRING_LEN];
        
        while ((res = ap_get_client_block(r, temp_buf, HUGE_STRING_LEN)) > 0) {
            *(temp_buf + res) = '\0';
            *(const char **)apr_array_push(a) = apr_pstrdup(temp_pool, temp_buf);
            /*this char will be deleted within this*/
        }
    } else {
#ifdef CSPACE_DEBUG
        dbg_printf("ERR: should_client_block false!");
#endif
        return FAIL;
    }

    *buf = apr_array_pstrcat(p, a, '\0'); /*use the request pool here*/
    
    apr_pool_destroy(temp_pool);

    return (res == OK);
}

static void set_header(char *key, char *val, void* table) {
    apr_table_t *t = (apr_table_t *)table;
    apr_table_set(t, strdup(key), strdup(val));/*is dup necessary?*/
}

static int find_uri(apr_array_header_t *a, char *uri) {
    int i;
    void *tmp;
    tmp = a->elts;
    for (i = 0; i < a->nelts; i++) {
        if (strcmp(*(char **)tmp, uri) == 0)
            return SUCC;
        tmp += a->elt_size;
    }

    return FAIL;
}

static int cspace_access_check(request_rec *r) {
    char *buf=NULL;
    int retval = OK;

#ifdef CSPACE_DEBUG
    dbg_printf(" ============================\n\n cspace_access_check: \n");
#endif

    cspace_cfg *svr_cfg = (cspace_cfg *)ap_get_module_config(
            r->server->module_config, &cspace_module);
  
    /*apr_table_set(r->headers_out, "Set-Cookie", "NAME=VALUE");*/
    /*apr_table_set(r->headers_out, "Location", "http://...url");*/
    /*apr_table_set(r->headers_out, "Location", "https://dummy:9090/login.html");
    return HTTP_MOVED_TEMPORARILY;*/

#ifdef CSPACE_DEBUG
    cfg_printf("\tsvr cfg", svr_cfg, "\n");
#endif

    /* To obtain per dir config:
     * cspace_cfg *dir_cfg = (cspace_cfg *)ap_get_module_config(r->per_dir_config,
     *        &cspace_module);
     */

#ifdef CSPACE_DEBUG
    dbg_printf("\turi: %s\n\n", r->uri);
#endif
    /*if we are not supposed to handle this req., hand it over to the next module
     * do this ASAP as we don't want our module to make other requests slow.
     */
    if (!find_uri(svr_cfg->use_cspace, r->uri)) return DECLINED;
    
    /*if we come this far the request is meant for us!
     * return HTTP_UNAUTHORIZED in case of an error with the request.
     */
    
    /*only https POST requests*/
    if (strcmp(ap_http_scheme(r), "https")) return HTTP_UNAUTHORIZED;
    if (r->method_number != M_POST) return HTTP_METHOD_NOT_ALLOWED;
   
    /*now we know that we are handling an https request*/
#ifdef CSPACE_DEBUG
    dbg_printf("https request\n");
#endif

    if (!cspace_read_request(r, &buf, &retval)) {
#ifdef CSPACE_DEBUG
        dbg_printf("ERR: request null! can't read?");
#endif        
        /*above: module should return that error code*/
        if(retval != OK) 
            return retval;
        else 
            return HTTP_UNAUTHORIZED;
    }
   
    if (buf) {
        int state = FAIL;
        process_ctx_t *ctx;
        
#ifdef CSPACE_DEBUG
        printf("\n<start>%s<end> \n", buf);
#endif

        cspace_decode_url((char *)buf);
        
        if (!(ctx = cspace_process_ctx_create((void *)(r->pool))))
            return HTTP_UNAUTHORIZED;
        
        state = (cspace_process_request(ctx, buf, svr_cfg->key_file, 
                        svr_cfg->epr, (void *)(r->headers_in), &set_header));
        if (!state)
            return HTTP_UNAUTHORIZED;

        /* at the moment we dont have to check if the state is 'authenticated' 
         * or 'pending', since we set that as a in header, so that the 
         * webapp has to hanlde it.
         */

        return OK;
    }

    return HTTP_UNAUTHORIZED; /*by this time we know that 
                                the request is for us*/
}

static void cspace_reg_hooks(apr_pool_t *p) {
#ifdef CSPACE_DEBUG
    dbg_printf("called: cspace_reg_hooks ******************\n");
#endif
    ap_hook_access_checker(cspace_access_check, NULL, NULL, APR_HOOK_FIRST);
}

module AP_MODULE_DECLARE_DATA cspace_module = {
    STANDARD20_MODULE_STUFF,
    NULL,/*cspace_cfg_perdir_create,*/
    NULL,/*cspace_cfg_perdir_merge,*/
    cspace_cfg_persvr_create,
    cspace_cfg_persvr_merge,
    cspace_cmds,
    cspace_reg_hooks
};
