/*
 * Copyright 2004,2005 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.solutions.identity.relyingparty.service;

import org.apache.axiom.om.OMAbstractFactory;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMFactory;
import org.apache.axiom.om.OMNamespace;
import org.apache.axis2.context.MessageContext;
import org.apache.axis2.context.ServiceContext;
import org.opensaml.XML;
import org.w3c.dom.Document;
import org.wso2.solutions.identity.relyingparty.RPConstants;
import org.wso2.solutions.identity.relyingparty.RelyingPartyException;
import org.wso2.solutions.identity.relyingparty.TokenVerifier;
import org.wso2.solutions.identity.relyingparty.Users;

import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilderFactory;

import java.io.ByteArrayInputStream;
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * This is the verfier service to verify decrypted token at a relying party
 */
public class VerifierService {

    /**
     * Verify a given decrupted xmlToken
     * 
     * @param request
     *            The verification request with the token to be verified and
     *            other information
     * @return If successfully verified then an OMElement with the claims
     *         available in the xmlToken. Otherwise an OMElement with faliure
     *         information.
     */
    public OMElement verify(OMElement request) throws RelyingPartyException {
        try {
            OMElement token = request.getFirstChildWithName(new QName(
                    RPConstants.NS, RPConstants.LN_TOKEN));
            String strVal = token.getFirstChildWithName(new QName(XML.SAML_NS, "Assertion")).toString();
            ByteArrayInputStream bais = new ByteArrayInputStream(strVal
                    .getBytes());

            // Convert to DOM
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            dbf.setNamespaceAware(true);
            Document doc = dbf.newDocumentBuilder().parse(bais);
            TokenVerifier tokenVerifier = new TokenVerifier();

            tokenVerifier.verifyDecryptedToken(doc.getDocumentElement());

            Hashtable attributeTable = tokenVerifier.getAttributeTable();

            // check whether the relying party is configured
            // to use self issued tokens
            if (isSelf(request)) {
                // In the case of the self issued tokens the relying party
                // should manage users
                Users users = Users.getInstance();
                
                ServiceContext serviceContext = MessageContext.getCurrentMessageContext().getServiceContext();
                String myEpr = serviceContext.getMyEPR().getAddress();
                String epr = myEpr.substring(0, myEpr.indexOf("RelyingPartyService")) + "UserRegistrationService/confirm?confString=";
                
//                AxisService service = MessageContext.getCurrentMessageContext()
//                        .getAxisServiceGroup().getService(
//                                "UserRegistrationService");
//                String epr = service.getEndpointURL();
                
                int state = users.handle(attributeTable, epr);
                if(state == Users.USER_REGISTERED) {
                    return getSuccessResponseElement(attributeTable); 
                } else {
                    return getConfirmationPendingResponseElement(attributeTable);
                }

            }

            return getSuccessResponseElement(attributeTable);
        } catch (Exception e) {
            e.printStackTrace();
            throw new RelyingPartyException(
                    RelyingPartyException.VERIFICATION_FAILURE, e);
        }

    }

    /**
     * @param attributeTable
     * @return
     */
    private OMElement getConfirmationPendingResponseElement(Hashtable attrTable) {
        OMFactory fac = OMAbstractFactory.getOMFactory();
        OMNamespace ns = fac.createOMNamespace(RPConstants.NS,
                RPConstants.PREFIX);
        OMElement resp = fac.createOMElement(RPConstants.LN_RESP, ns);
        resp.addAttribute(RPConstants.ATTR_STATE,
                RPConstants.STATE_NEW_USER_CONFIRMATION_PENDING, null);
        Enumeration keysEnum = attrTable.keys();
        while (keysEnum.hasMoreElements()) {
            String attrName = (String) keysEnum.nextElement();
            OMElement prop = fac.createOMElement(
                    RPConstants.LN_PROPERTY, ns);
            prop.setText((String) attrTable.get(attrName));
            prop.addAttribute(fac.createOMAttribute(
                    RPConstants.ATTR_NAME, null, attrName));
            resp.addChild(prop);
        }

        return resp;
    }

    /**
     * Figure out whether we are processing a self issued token or not.
     * 
     * @param request
     *            The incoming verification request
     * @return
     */
    private boolean isSelf(OMElement request) {
        OMElement issuer = request.getFirstChildWithName(new QName(
                RPConstants.NS, RPConstants.LN_ISSUER));
        if (issuer != null && issuer.getText() != null) {
            return RPConstants.ISSUER_SELF.equals(issuer.getText()
                    .trim());
        }

        return false;
    }

    /**
     * Create the response element with the given attribute table
     * 
     * @param attrTable
     *            table of attributes to be added to the response
     * @return The response element to be returned
     */
    private OMElement getSuccessResponseElement(Hashtable attrTable) {
        OMFactory fac = OMAbstractFactory.getOMFactory();
        OMNamespace ns = fac.createOMNamespace(RPConstants.NS,
                RPConstants.PREFIX);
        OMElement resp = fac.createOMElement(RPConstants.LN_RESP, ns);
        resp.addAttribute(RPConstants.ATTR_STATE,
                RPConstants.STATE_AUTHENTICATED, null);
        Enumeration keysEnum = attrTable.keys();
        while (keysEnum.hasMoreElements()) {
            String attrName = (String) keysEnum.nextElement();
            OMElement prop = fac.createOMElement(
                    RPConstants.LN_PROPERTY, ns);
            prop.setText((String) attrTable.get(attrName));
            prop.addAttribute(fac.createOMAttribute(
                    RPConstants.ATTR_NAME, null, attrName));
            resp.addChild(prop);
        }

        return resp;
    }

}
